/*
 * gscan -- User Driver for Genius ColorPage-CS Desktop Scanner
 *  
 * Copyright (C) 1997 Rildo Pragana <rpragana@acm.org>     
 * S-mail: Caixa Postal 7440, Recife Brazil 50,630-970
 * Based on the software muscan by Torsten Eichner
 *
 * This is free software, following Free Software Foundation's GNU
 * philosophy. Please read GNU Copyright statements.
 * You may freely modify, use in commercial or personal software, but
 * please, acknowledge its author. You may also ask me for changes and 
 * improvements. Feel free to contact me.
 */

#define GSCANVER "Gscan Version 0.0.1 (C) Copyright 1997, Rildo Pragana"

#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include </usr/src/linux/drivers/scsi/sg.h>

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define MAXCOLS 3000
#define MAXROWS 43000
#define MAXHIST 500
#define PBM_WHITE 0
#define PBM_BLACK 1

static void open_scanner (void);
static void close_scanner (void);
static void signal_handler (int);
static int scsi_cmd (unsigned char *, int, unsigned char *, int);
static void usage( void );
static void scan_image( void );
static void write_image ( void );

static char *scanner_device = DEVICE;
static int scanner_fd = -1;

static int outf=STDOUT_FILENO; /* default - stdout */
static long left=0;
static long top=0;
static long width=0;
static long height=0;
static int xres=0x19;
static int yres=0x19;
static int brightness=0;
static int contrast=128;

static int mem_size = 4096;
static int debug = 0;

static unsigned char result[256];
#define MAXSGRD (SG_BIG_BUFF - 1024)
static unsigned char *image;

static unsigned char inquiry[] =
  {0x12, 0x00, 0x00, 0x00, 0xff, 0x00};
static unsigned char testunit[] =
  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static unsigned char req_sense[] =
  {0x03, 0x00, 0x00, 0x00, 0x04, 0x00};
static unsigned char scan[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
static unsigned char readcmd[] =
	{0x28,0, 0x00, 0, 0, 0x00, 0x00,0x10,0x00, 0};
static unsigned char getbufstat[] =
	{0x34,1,0,0,0,0,0,0x00,0x10,0};

static unsigned char win[] =
		{0x24,0,0,0,0,0, 0x00,0x00,0x30, 0, /* command header */
		 0,0x2e,0,0,0,0, 0x00,0x28,         /* window parameter block */
		 0,0,								/* Window Descriptor (ID,reserved) */
		 0x00,0x19, 0x00,0x19,              /* [20] X,Y resolution */
 		 0x00,0x00,0x00,0x00,				/* [24] Upper left X 1'' */
 		 0x00,0x00,0x00,0x00,				/* [28] Upper left Y 4'' */
 		 0x00,0x00,0x00,0xc8,				/* [32] width */
 		 0x00,0x00,0x01,0x13,				/* [36] height */
 		 0x00,0x80,0x00,					/* brightness, threshold, contrast */
 		 0x05,0x08,0x00,0x00,0x25,0,0,
 		 0,0,0,1,0,0,0,0};
static unsigned char nullwin[] = /* this window is to home the carriage */
		{0x24,0,0,0,0,0, 0x00,0x00,0x30, 0, /* command header */
		 0,0x2e,0,0,0,0, 0x00,0x28,         /* window parameter block */
		 0,0,								/* Window Descriptor (ID,reserved) */
		 0x00,0x19, 0x00,0x19,              /* [20] X,Y resolution */
 		 0x00,0x00,0x00,0x00,				/* [24] Upper left X 1'' */
 		 0x00,0x00,0x00,0x00,				/* [28] Upper left Y 4'' */
 		 0x00,0x00,0x00,0xc8,				/* [32] width */
 		 0x00,0x00,0x00,0x00,				/* [36] height */
 		 0x00,0x80,0x00,					/* brightness, threshold, contrast */
 		 0x05,0x08,0x00,0x00,0x25,0,0,
 		 0,0,0,1,0,0,0,0};

#define ERROR(string...) { fprintf(stderr, "ERROR[%s:%d]: ", __FILE__, __LINE__); fprintf(stderr, ##string); exit(-1); }
#define TRACE(lvl, string...) { if (debug >= lvl) { fprintf(stderr, "TRACE[%s:%d]: ", __FILE__, __LINE__); fprintf(stderr, ##string);fflush(stderr); } }

/*******************************************
 * Signal handler                          *
 *******************************************/
static void signal_handler ( int signal )
{
  if (scanner_fd != -1)
      close_scanner ();
  exit (1);
}

/*******************************************
 * Issue a SCSI command                    *
 *******************************************/
static int scsi_cmd( unsigned char *src, int src_size,
	unsigned char *dest, int dest_size ) {
#ifndef SG_BIG_BUFF
  unsigned char scsi_buffer[4096 + sizeof (struct sg_header)];
#else
  unsigned char scsi_buffer[SG_BIG_BUFF];
#endif
  struct sg_header hdr;
  int i, count;

  TRACE (3, "issue SCSI command 0x%02x (expected replylen: %d+%d=%d)\n", src[0], dest_size, sizeof (struct sg_header), dest_size + sizeof (struct sg_header));

#ifdef SG_BIG_BUF
  if (dest_size > SG_BIG_BUF - sizeof (struct sg_header))
      ERROR ("buffer overflow: expected %d bytes, buffer is %d bytes\n", count, SG_BIG_BUF - sizeof (struct sg_header));
#endif

  hdr.pack_len = src_size + sizeof (struct sg_header);
  hdr.reply_len = dest_size + sizeof (struct sg_header);

  for (i = 0; i < sizeof (struct sg_header); i++)
    scsi_buffer[i] = ((char *) &hdr)[i];

  for (; i < hdr.pack_len; i++)
    scsi_buffer[i] = src[i - sizeof (struct sg_header)];
  
  	if (write (scanner_fd, scsi_buffer, hdr.pack_len) != hdr.pack_len)
   		ERROR ("error on write: %s\n", strerror (errno));

  	if ((count = read (scanner_fd, scsi_buffer, dest_size + sizeof (struct sg_header))) < 0)
    	ERROR ("error on read: %s\n", strerror (errno));
  
  for (i = 0; i < sizeof (struct sg_header); i++)
    ((char *) &hdr)[i] = scsi_buffer[i];

  if (hdr.result != 0)
    ERROR ("SCSI error (%02x): (cmd %02x) sense 0x%02x 0x%02x 0x%02x...\n", hdr.result, src[0], hdr.sense_buffer[0], hdr.sense_buffer[1], hdr.sense_buffer[2]);

  TRACE (3, "SCSI: read %4d packlen %4d replylen %d result 0x%02x\n", count, hdr.pack_len, hdr.reply_len, hdr.result);
  TRACE (3, "SCSI: sense %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n", hdr.sense_buffer[0], hdr.sense_buffer[1], hdr.sense_buffer[2], hdr.sense_buffer[3], hdr.sense_buffer[4], hdr.sense_buffer[5], hdr.sense_buffer[6], hdr.sense_buffer[7], hdr.sense_buffer[8], hdr.sense_buffer[9], hdr.sense_buffer[10], hdr.sense_buffer[11], hdr.sense_buffer[12], hdr.sense_buffer[13], hdr.sense_buffer[14], hdr.sense_buffer[15]);

  TRACE (4, "SCSI data (BEGIN)\n");

  for (; i < count; i++)
    {
      dest[i - sizeof (struct sg_header)] = scsi_buffer[i];
      if ((src[0] != 0x08 && debug > 2) || (debug > 3))
      	fprintf (stderr, "%4x:%02x(%c)%s", i - sizeof (struct sg_header), scsi_buffer[i], scsi_buffer[i] >= 32 && scsi_buffer[i] < 127 ? scsi_buffer[i] : ' ', (i - sizeof (struct sg_header) + 1) % 8 == 0 ? "\n" : "");
    }

  TRACE (4, "SCSI data (END)\n");

  return count - sizeof (struct sg_header);
}


/*******************************************
 *  Set Window                             *
 *******************************************/
static void set_window ( void ) {
	win[20] = xres/256; win[21] = xres%256;
	win[22] = yres/256; win[23] = yres%256;
	win[25] = left/65536; win[26] = (left/256)%256; win[27] = left%256;
	win[29] = top/65536; win[30] = (top/256)%256; win[31] = top%256;
	win[33] = width/65536; win[34] = (width/256)%256; win[35] = width%256;
	win[37] = height/65536; win[38] = (height/256)%256; win[39] = height%256;
	win[40] = brightness;
	win[42] = contrast;

	TRACE(1,"Window: xres=%02x%02x, yres=%02x%02x, width=%02x%02x%02x, height=%02x%02x%02x\n",
		win[20],win[21],win[22],win[23],
		win[33],win[34],win[35],win[37],win[38],win[39]);

	scsi_cmd(win,66,result,256);
} 		 

/*******************************************
 *  Home scanner                           *
 *******************************************/

static void home_scanner ( void ) {
	 
	scsi_cmd (testunit, 6, NULL, 0);
	scsi_cmd(nullwin,66,result,256);
    scsi_cmd (scan, 6, NULL,0); /* comando SCAN */
} 		 

 
/*******************************************
 *  Read scanner data                      *
 *******************************************/

static void scan_read ( void ) {
	unsigned int available,filled,readno;
	unsigned char buf_status[0x10];
	long bufoff;
	int totalline;
	
	filled=1;
	totalline=width*3;
	bufoff=0;
	while (filled>0) {
		available = 0;
		while (available == 0) {
			scsi_cmd(getbufstat,10,buf_status,0x10);
			available = (buf_status[9]*0x100 + buf_status[10]) * 0x100 +
				buf_status[11];
			filled = buf_status[12]*0x100+buf_status[13];
		}
		while (available) {
			readno = available;    
			if (readno > MAXSGRD) readno = MAXSGRD;
			readcmd[6] = readno / 65536;
			readcmd[7] = (readno / 256) % 256;
			readcmd[8] = readno % 256;
			scsi_cmd(readcmd,10,image+bufoff,readno);
			bufoff += readno;
			available -= readno;
		}	
		TRACE(2,"Genius: filled lines: %d\n",filled);
	}
} 		 
 
/*******************************************
 * Write line                              *
 *******************************************/
static void write_image ( void )
{
  	long x,y,offset;
  	unsigned char pixel[3];
  	for (y = 0; y < (height*3); y+=3) {
  		offset = y * width;
		TRACE(2,"Writing %ld line\n",y/3);
		for (x = 0; x < width; x++) {
		    pixel[0] = image[x+offset];					/* R channel */
		    pixel[1] = image[x + offset + width];		/* G channel */
	        pixel[2] = image[x + offset + width * 2];	/* B channel */
			write(outf,pixel,3);
	  	}
	}
}

/*******************************************
 * Open scanner                            *
 *******************************************/
static void open_scanner ( void )
{
#ifdef SG_BIG_BUFF
  union
    {
      unsigned char tmp[SG_BIG_BUFF + sizeof (struct sg_header)];
      struct sg_header hdr;
    }
  sgb;
#endif
  int i;
  int valid_size;

  TRACE (1, "opening scanner <%s>\n", scanner_device);
	  if ((scanner_fd = open (scanner_device, O_RDWR | O_EXCL)) < 0)
    	ERROR ("cannot open scanner <%s> due to: %s\n", scanner_device, strerror (errno));
  scsi_cmd (inquiry, 6, result, 256);

#ifdef SG_BIG_BUFF
  /* try to determine maximum buffer size */
  for (valid_size = 4096; valid_size < SG_BIG_BUFF; valid_size += 1024)
    {
      sgb.hdr.pack_len = 6 + sizeof (struct sg_header);
      sgb.hdr.reply_len = valid_size;

      for (i = 0; i < 6; i++)
	sgb.tmp[i + sizeof (struct sg_header)] = inquiry[i];

      if (write (scanner_fd, &sgb, 6 + sizeof (struct sg_header)) < 0)
	 	 break;

      read (scanner_fd, &sgb, sgb.hdr.reply_len + sizeof (struct sg_header));
      mem_size = valid_size;
    }

  TRACE (1, "new mem_size is %d\n", mem_size);
#endif
}

/*******************************************
 * Close scanner                           *
 *******************************************/
static void close_scanner ( void )
{
  TRACE (1, "closing scanner <%s>\n", scanner_device);
  	close (scanner_fd);
  scanner_fd = -1;
}

/*******************************************
 * Send Gamma                              *
 *******************************************/
static void send_gamma( void ) {
	unsigned char buf[0x30a];
	int i;
	
	for (i=0;i<10;i++) buf[i]=0;
	buf[0] = 0x2a;
	buf[2] = 3;
	buf[5] = 2;
	buf[7] = 3;
	for (i=0;i<0x300;i++)
		buf[i+10] = i & 0xff;
	scsi_cmd(buf,0x30a,NULL,0);
}

/*******************************************
 * Scan image GENIUS ColorPage-CS          *
 *******************************************/
static void scan_image ( void )
{
  
  unsigned char resbuf[256];
  TRACE (1, "scan image\n");

  {
    scsi_cmd (inquiry, 6, resbuf, 256);
	scsi_cmd (req_sense,6, resbuf, 256);
	scsi_cmd (testunit, 6, NULL, 0);

    set_window();
    send_gamma();
	set_window();	
    scsi_cmd (scan, 6, NULL,0); /* comando SCAN */
    sleep(18); /* this is needed because of a bug in the low-level driver!? */
	scan_read();
  }
  
}

static void usage( void ) {
	fprintf(stderr,"\n%s\n",GSCANVER);
	fprintf(stderr,"usage: gscan [noltwhxyrdbc]\n\n");
	fprintf(stderr,"\t-?          -- show this usage info\n");
	fprintf(stderr,"\t-o <file>   -- outputs to file instead of stdout\n");
	fprintf(stderr,"\t-l <margin> -- lateral (right) margin\n");
	fprintf(stderr,"\t-t <margin> -- top margin\n");
	fprintf(stderr,"\t-w <size>   -- width of object\n");
	fprintf(stderr,"\t-h <size>   -- height of object\n");
	fprintf(stderr,"\t-x <dpi>    -- x resolution (dpi)\n");
	fprintf(stderr,"\t-y <dpi>    -- y resolution (dpi)\n");
	fprintf(stderr,"\t-r <dpi>    -- x and y resolution (dpi)\n");
	fprintf(stderr,"\t-d <level>  -- debugging level of output\n");
	fprintf(stderr,"\t-b <brightness>\n");
	fprintf(stderr,"\t-c <contrast>\n");

	fprintf(stderr,"\n\toutput goes as default to stdout\n");
	fprintf(stderr,"\tdefault resolution is 25 dpi (preview mode)\n");
	fprintf(stderr,"\tdefault width and height is maximum (8.5x11.7in)\n");
	fprintf(stderr,"\tunits are in 0.1 inch\n");
	fprintf(stderr,"\tbrightness and contrast doesn't work yet :(\n");
	exit(0);
}

/*******************************************
 * Main procedure                          *
 *******************************************/
void main( int argc, char *argv[] ) 
{
  int c;
  char s[200];

  while ((c = getopt(argc, argv, "?b:c:no:l:t:w:h:x:y:d:r:")) != EOF) {
    switch (c) {
    case '?': usage(); break;
    case 'o': outf=open(optarg,O_CREAT|O_TRUNC|O_RDWR); break;
	case 'l': left = atol(optarg); break;
	case 't': top = atol(optarg); break;
	case 'w': width = atol(optarg); break;
	case 'h': height = atol(optarg); break;
	case 'x': xres = atoi(optarg); break;
	case 'y': yres = atoi(optarg); break;
	case 'r': xres = yres = atoi(optarg); break;
	case 'd': debug = atoi(optarg); break;
	case 'b': brightness = atoi(optarg); break;
	case 'c': contrast = atoi(optarg); break;
	default: fprintf(stderr,"Invalid switch\n"); exit(1);
 	}
  }
  /* convert parameters to inches */
  if (width) 
  	width = (width * xres) / 10;
  else
  	width = (85 * xres) / 10; /* default width 8.5in */
  if (height)
  	height = (height * yres) / 10;
  else
  	height = (117 * yres) / 10; /* default height 11in */
  if (left) left = (left * xres) / 10;
  if (top) top = (top * yres) / 10;
  
  /* make sure that window is inside limits */
  width = min( width, (85 * xres) / 10 - left );
  height = min( height, (117 * yres) / 10 - top );
 
  TRACE(1,"Dimensions: %ld x %ld pixels\n",width,height); 
  /* prepare image buffer */
  sprintf( s,"P6\n# gscan generated image \n%d %d\n255\n",(int)width,(int)height );
  write( outf,s,strlen(s) );
  TRACE(2,"Alloc %ld byte for image buffer\n",width*height*3);
  if ((image = malloc( width * height * 3 ))==NULL) { 
  	close(outf);
	fprintf(stderr,"No memory for the image\n");
  	exit(1);
  }

  signal (SIGHUP, signal_handler);
  signal (SIGINT, signal_handler);
  signal (SIGQUIT, signal_handler);
  signal (SIGPIPE, signal_handler);
  signal (SIGALRM, signal_handler);
  signal (SIGTERM, signal_handler);
  signal (SIGXCPU, signal_handler);
  signal (SIGXFSZ, signal_handler);
  signal (SIGVTALRM, signal_handler);
  signal (SIGPROF, signal_handler);
  signal (SIGUSR1, signal_handler);
  signal (SIGUSR2, signal_handler);

  open_scanner (); 
  scan_image ();
  home_scanner();
  write_image();
  if (outf!=STDOUT_FILENO) close(outf);
  free(image);
  close_scanner ();
}

/* End of file */
